"""Implements twitter."""

import argparse
import logging

import tweepy

from utils import timex

MAX_LEN_TWEET = 280
MAX_MEDIA_FILES = 4
log = logging.getLogger('twitter-wrapper')
logging.basicConfig(level=logging.INFO)


def _update_status(api, tweet_text, media_ids, in_reply_to_status_id):
    if len(media_ids) > 0 and in_reply_to_status_id:
        return api.update_status(
            tweet_text,
            media_ids=media_ids,
            in_reply_to_status_id=in_reply_to_status_id,
        )

    if len(media_ids) > 0:
        return api.update_status(tweet_text, media_ids=media_ids)

    if in_reply_to_status_id:
        return api.update_status(
            tweet_text, in_reply_to_status_id=in_reply_to_status_id
        )

    return api.update_status(tweet_text)


def _upload_media(api, image_files):
    media_ids = []
    for image_file in image_files:
        media_id = api.media_upload(image_file).media_id
        media_ids.append(media_id)
        log.info(
            'Uploaded status image %s to twitter as %s',
            image_file,
            media_id,
        )
    return media_ids


def _update_profile_description(api):
    date_with_timezone = timex.format_current_date_with_timezone()
    description = 'Automatically updated at {date_with_timezone}'.format(
        date_with_timezone=date_with_timezone
    )
    api.update_profile(description=description)
    log.info('Updated profile description to: %s', description)


class Twitter:
    """Implements Twitter wrapper."""

    def __init__(
        self,
        twtr_api_key,
        twtr_api_secret_key,
        twtr_access_token,
        twtr_access_token_secret,
    ):
        """Construct Twitter."""
        if twtr_api_key:
            auth = tweepy.OAuthHandler(twtr_api_key, twtr_api_secret_key)
            auth.set_access_token(twtr_access_token, twtr_access_token_secret)
            self.api = tweepy.API(auth)
            api_me = self.api.me()
            log.info(
                'Created Twitter for @%s (%s).',
                api_me.screen_name,
                api_me.id_str,
            )
        else:
            log.error('Missing twitter API Key etc. Cannot create Twitter.')
            self.api = None

    @staticmethod
    def from_args(description=''):
        """Construct Twitter from Args."""
        parser = argparse.ArgumentParser(description=description)
        for twtr_arg_name in [
            'twtr_api_key',
            'twtr_api_secret_key',
            'twtr_access_token',
            'twtr_access_token_secret',
        ]:
            parser.add_argument(
                '--' + twtr_arg_name,
                type=str,
                required=False,
                default=None,
            )
        args = parser.parse_args()
        return Twitter(
            args.twtr_api_key,
            args.twtr_api_secret_key,
            args.twtr_access_token,
            args.twtr_access_token_secret,
        )

    def tweet(
        self,
        tweet_text,
        status_image_files=None,
        update_user_profile=False,
        profile_image_file=None,
        banner_image_file=None,
        in_reply_to_status_id=None,
    ):
        """Tweet."""
        if status_image_files is None:
            status_image_files = []

        log.info('tweet_text: %s', tweet_text)
        log.info('status_image_files: %s', str(status_image_files))
        log.info('update_user_profile: %s', str(update_user_profile))
        log.info('profile_image_file: %s', str(profile_image_file))
        log.info('banner_image_file: %s', str(banner_image_file))

        tweet_text_for_length_list = []
        for tweet_text_line in tweet_text.split('\n'):
            if 'http' not in tweet_text_line:
                tweet_text_for_length_list.append(tweet_text_line)
        n_tweet_text = len('\n'.join(tweet_text_for_length_list))

        if n_tweet_text > MAX_LEN_TWEET:
            log.error(
                'Tweet text is too long (%d chars). Not tweeting.',
                n_tweet_text,
            )
            return None
        else:
            log.info('Tweet Length: %d', n_tweet_text)

        if len(status_image_files) > MAX_MEDIA_FILES:
            log.warning('Too many (%d) status image files. Truncating.')
            status_image_files = status_image_files[:MAX_MEDIA_FILES]

        if not self.api:
            log.error('Missing API. Cannot tweet')
            return None

        media_ids = _upload_media(self.api, status_image_files)
        update_status_result = _update_status(
            self.api, tweet_text, media_ids, in_reply_to_status_id
        )

        if update_user_profile:
            _update_profile_description(self.api)

        if profile_image_file:
            self.api.update_profile_image(profile_image_file)
            log.info('Update profile image to %s', profile_image_file)

        if banner_image_file:
            self.api.update_profile_banner(banner_image_file)
            log.info('Update profile banner image to %s', banner_image_file)

        return update_status_result



"""Geo-Spatial Utils."""
from math import atan2, cos, pi, sin, sqrt

from area import area

EARTH_RADIUS = 6373.0
QUANTUMS = 1_000_000
LAT_LNG_COLOMBO = [6.9271, 79.8612]
LAT_LNG_KANDY = [7.2906, 80.6337]


def parse_latlng(latlng_str):
    """Parse latlng string."""
    latlng_str = latlng_str.replace('°', '')
    lat_sign = 1
    if 'N' in latlng_str:
        latlng_str = latlng_str.replace('N', '')
    elif 'S' in latlng_str:
        latlng_str = latlng_str.replace('S', '')
        lat_sign = -1

    lng_sign = 1
    if 'E' in latlng_str:
        latlng_str = latlng_str.replace('E', '')
    elif 'W' in latlng_str:
        latlng_str = latlng_str.replace('W', '')
        lng_sign = -1

    lat_str, lng_str = latlng_str.split(',')
    return (float)(lat_str) * lat_sign, (float)(lng_str) * lng_sign


def deg_to_rad(deg):
    """Convert degrees to radians."""
    deg_round = round(deg * QUANTUMS, 0) / QUANTUMS
    return deg_round * pi / 180


def get_distance(latlng1, latlng2):
    """Get distance between two points."""
    lat1, lng1 = latlng1
    lat2, lng2 = latlng2

    lat1 = deg_to_rad(lat1)
    lng1 = deg_to_rad(lng1)
    lat2 = deg_to_rad(lat2)
    lng2 = deg_to_rad(lng2)

    dlat = lat2 - lat1
    dlng = lng2 - lng1

    a_var = (sin(dlat / 2)) ** 2 + cos(lat1) * cos(lat2) * (sin(dlng / 2)) ** 2
    c_var = 2 * atan2(sqrt(a_var), sqrt(1 - a_var))
    return EARTH_RADIUS * c_var


def get_area(lnglat_list_list):
    """Find the area of a lnglat list list."""

    def get_area_for_lnglat_list(lnglat_list):
        obj = {
            'type': 'Polygon',
            'coordinates': [lnglat_list],
        }
        return area(obj) / 1000_000

    return sum(list(map(get_area_for_lnglat_list, lnglat_list_list)))



"""JSON utils."""
import base64
import json

from geopandas.geodataframe import GeoDataFrame
from pandas import read_pickle
from pandas.core.frame import DataFrame
from shapely.geometry import mapping, shape
from shapely.geometry.base import BaseGeometry

from utils import filex


def read(file_name):
    """Read JSON from file.

    Args:
        file_name (str): file name

    Returns:
        Parsed JSON data

    """
    return json.loads(filex.read(file_name))


def write(file_name, data):
    """Write data as JSON to file.

    Args:
        file_name (str): file name
        data: data as serializable object

    """
    filex.write(file_name, json.dumps(data, indent=2))


def serialize(data):
    if isinstance(data, bytes):
        return {
            'type': 'bytes',
            'data': base64.b64encode(data).decode('ascii'),
        }
    if isinstance(data, BaseGeometry):
        return {
            'type': 'BaseGeometry',
            'data': mapping(data),
        }

    if isinstance(data, GeoDataFrame):
        tmp_file = filex.get_tmp_file()
        return {
            'type': 'GeoDataFrame',
            'data': data.to_pickle(tmp_file),
            'pickle_file': tmp_file,
        }
    if isinstance(data, DataFrame):
        tmp_file = filex.get_tmp_file()
        return {
            'type': 'DataFrame',
            'data': data.to_pickle(tmp_file),
            'pickle_file': tmp_file,
        }

    return {
        'type': None,
        'data': data,
    }


def deserialize(data):
    data_type = data['type']
    data_data = data['data']

    if data_type == 'bytes':
        return base64.b64decode(data_data)
    if data_type == 'BaseGeometry':
        return shape(data_data)

    if data_type == 'GeoDataFrame':
        pickle_file = data['pickle_file']
        return read_pickle(pickle_file)
    if data_type == 'DataFrame':
        pickle_file = data['pickle_file']
        return read_pickle(pickle_file)

    return data_data



"""Map-Reduce utils."""

import concurrent.futures

DEFAULT_MAX_THREADS = 4


def _run_parallel(workers, max_threads=DEFAULT_MAX_THREADS):
    """Run workers in parallel."""
    with concurrent.futures.ThreadPoolExecutor(max_threads) as thread_pool:
        future_list = []
        for worker in workers:
            future = thread_pool.submit(worker)
            future_list.append(future)

        for future in future_list:
            future.done()

        output_list = []
        for future in future_list:
            output_list.append(future.result())
        return output_list


def map_parallel(func_map, params_list, max_threads=DEFAULT_MAX_THREADS):
    """Run list(map(...)) in parallel.

    Args:
        func_map (function): Mapper function
        params_list (list): Params to be mapped
        max_threads (int, optional): Maximum parallel threads.
            DEFAULT_MAX_THREADS = 4

    .. code-block:: python

        >>> from utils import mr
        >>> print(mr.map_parallel(lambda x: x ** 2, [1, 2, 3, 4]))
        [1, 4, 9, 16]

    """

    def get_worker(params):
        def worker():
            return func_map(params)

        return worker

    workers = list(
        map(
            lambda params: get_worker(params=params),
            params_list,
        )
    )
    return _run_parallel(workers, max_threads)



"""Utils for caching functions."""
import json
from functools import wraps

from utils.cache_impl import _Cache

CACHE_DIR = '/tmp/cache'
CACHE_DEFAULT_TIMEOUT = 60


def cache(cache_name, timeout=CACHE_DEFAULT_TIMEOUT):
    """Wrap class Cache as decorator."""

    def cache_inner(func):
        @wraps(func)
        def cache_inner_inner(*args, **kwargs):
            cache_key = json.dumps(
                {
                    'cache_name': cache_name,
                    'function_name': func.__name__,
                    'kwargs': kwargs,
                    'args': args,
                }
            )

            def fallback():
                return func(*args, **kwargs)

            return _Cache(cache_name, timeout).get(cache_key, fallback)

        return cache_inner_inner

    return cache_inner
